#ifndef __PPA_API_MFE_H__20100427_1703
#define __PPA_API_MFE_H__20100427_1703
/*******************************************************************************
**
** FILE NAME    : ppa_api_qos.h
** PROJECT      : PPA
** MODULES      : PPA API (PPA QOS  APIs)
**
** DATE         : 27 April 2010
** AUTHOR       : Shao Guohua
** DESCRIPTION  : PPA QOS APIs
**                File
** COPYRIGHT    : Copyright (c) 2017 Intel Corporation
**
**   For licensing information, see the file 'LICENSE' in the root folder of
**   this software module.
**
** HISTORY
** $Date                $Author         $Comment
** 16 Dec 2009  Shao Guohua             Initiate Version
*******************************************************************************/

/*! \file ppa_api_qos.h
\brief This file contains es.
provide PPA power management API.
*/

/** \addtogroup  PPA_API_QOS */
/*@{*/

/*
* ####################################
*              Definition
* ####################################
*/


/*
* ####################################
*              Data Type
* ####################################
*/


/*
* ####################################
*             Declaration
* ####################################
*/
/*typedef int (*PPA_QOS_CLASS2PRIO_CB)(int32_t port_id, struct net_device *netif, uint8_t *class2prio);*/
typedef struct qos_queue_list_item {
        struct qos_queue_list_item                      *next;
        PPA_ATOMIC                                      count;
        char                                    ifname[16];/*!< Interface name on which the Queue is modified*/
        char                                    dev_name[16];/*!< Interface name on which the Queue is modified*/
        char                                    tc_map[MAX_TC_NUM];/*!<Which all Traffic Class(es) map to this Queue*/
        uint8_t                                 tc_no;/*!<Nunber of Traffic Class(es) map to this Queue*/
        uint8_t                                 intfId_en;/*!<Enable/Disable for flow Id + tc bits used for VAP's & Vlan interfaces*/
        uint16_t                                        intfId;/*!< flow Id + tc bits used for VAP's & Vlan interfaces*/
        uint32_t                                enable; /*!< Whether Queue is enabled */
        int32_t                                 weight; /*!< WFQ Weight */
        int32_t                                 priority; /*!< Queue Priority or Precedence. Start from 1-16, with 1 as highest priority */
        int32_t                                 qlen; /*!< Length of Queue in bytes */
        int32_t                                 queue_num; /*!< Logical queue number added will be filled up in this field */
        int32_t                                 shaper_num; /*!< Logical shaper number added will be filled up in this field */
        uint32_t                                portid;/*!< Portid*/
        PPA_QOS_DROP_CFG                                drop; /*!< Queue Drop Properties */
        uint32_t                                flags;/*!< Flags for future use*/
        uint32_t                                p_entry;/*!< Physical index of the Queue created returned by the Hal*/
        void                                    *caps_list;/*!< List of capabilities for the specific Queue*/
        uint16_t                                num_caps;/*!< Maximum number of capabilities set for specific Queue*/

} QOS_QUEUE_LIST_ITEM;

#if defined(CONFIG_PPA_VAP_QOS_SUPPORT) && defined(CONFIG_SOC_GRX500)
typedef struct qos_logical_if_list_item {
        struct qos_logical_if_list_item *next;
        char logical_ifname[16];
        int32_t port_id;
        int32_t subif_id;
        uint8_t flowId_en;
        int32_t flowId;

} QOS_LOGICAL_IF_LIST_ITEM;
#endif

#ifdef CONFIG_PPA_QOS
extern int32_t ppa_ioctl_get_qos_status(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_get_qos_qnum(uint32_t portid, uint32_t flag);
extern int32_t ppa_ioctl_get_qos_qnum(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_get_qos_mib(uint32_t portid, uint32_t queueid, PPA_QOS_MIB *mib, uint32_t flag);
extern int32_t ppa_ioctl_get_qos_mib(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);

extern int32_t ppa_set_qos_wfq(uint32_t portid, uint32_t queueid, uint32_t weight_level, uint32_t flag);
extern int32_t ppa_get_qos_wfq(uint32_t portid, uint32_t queueid, uint32_t *weight_level, uint32_t flag);
extern int32_t ppa_reset_qos_wfq(uint32_t portid, uint32_t queueid, uint32_t flag);
extern int32_t ppa_set_ctrl_qos_wfq(uint32_t portid,  uint32_t f_enable, uint32_t flag);
extern int32_t ppa_get_ctrl_qos_wfq(uint32_t portid,  uint32_t *f_enable, uint32_t flag);

extern int32_t ppa_ioctl_add_qos_queue(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_modify_qos_queue(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_delete_qos_queue(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_qos_init_cfg(unsigned int cmd);
#if defined(CONFIG_SOC_GRX500) && CONFIG_SOC_GRX500
extern int32_t ppa_ioctl_add_class_rule(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_mod_class_rule(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_del_class_rule(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_class_rule(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_set_qos_ingress_group(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_qos_ingress_group(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
#endif

extern int32_t ppa_ioctl_mod_subif_port_config(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_add_wmm_qos_queue(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_delete_wmm_qos_queue(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_set_ctrl_qos_wfq(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_ctrl_qos_wfq(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_set_qos_wfq(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_qos_wfq(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_reset_qos_wfq(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);

extern int32_t ppa_set_ctrl_qos_rate(uint32_t portid,  uint32_t f_enable, uint32_t flag);
extern int32_t ppa_get_ctrl_qos_rate(uint32_t portid,  uint32_t *f_enable, uint32_t flag);
#if defined(MBR_CONFIG) && MBR_CONFIG
extern int32_t ppa_set_qos_rate(char *ifname, uint32_t portid, uint32_t queueid, int32_t shaperid, uint32_t rate, uint32_t burst, uint32_t flag, int32_t hal_id);
extern int32_t ppa_get_qos_rate(uint32_t portid, uint32_t queueid, int32_t *shaperid, uint32_t *rate, uint32_t *burst, uint32_t flag);
#else
extern int32_t ppa_set_qos_rate(char *ifname, uint32_t portid, uint32_t queueid, uint32_t rate, uint32_t burst, uint32_t flag, int32_t hal_id);
extern int32_t ppa_get_qos_rate(uint32_t portid, uint32_t queueid, uint32_t *rate, uint32_t *burst, uint32_t flag);
#endif
#if defined(CONFIG_PPA_HAL_SELECTOR) && CONFIG_PPA_HAL_SELECTOR
extern int32_t ppa_reset_qos_rate( char *ifname, char *dev_name, uint32_t portid, int32_t queueid, int32_t shaperid, uint32_t flag, int32_t hal_id);
#else
extern int32_t ppa_reset_qos_rate(uint32_t portid, int32_t queueid, uint32_t flag);
#endif
#if defined(MBR_CONFIG) && MBR_CONFIG
extern int32_t ppa_set_qos_shaper(int32_t shaperid, uint32_t rate, uint32_t burst, PPA_QOS_ADD_SHAPER_CFG *, uint32_t flags, int32_t hal_id);
extern int32_t ppa_get_qos_shaper(int32_t shaperid, uint32_t *rate, uint32_t *burst, uint32_t flag);
extern int32_t ppa_ioctl_set_qos_shaper(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_qos_shaper(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
#endif
extern int32_t ppa_ioctl_set_ctrl_qos_rate(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_ctrl_qos_rate(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_set_qos_rate(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_reset_qos_rate(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);
extern int32_t ppa_ioctl_get_qos_rate(unsigned int cmd, unsigned long arg, PPA_CMD_DATA *cmd_info);

#if defined(CONFIG_PPA_VAP_QOS_SUPPORT) && defined(CONFIG_SOC_GRX500)
extern QOS_LOGICAL_IF_LIST_ITEM *g_qos_logical_if;
extern PPA_LOCK g_qos_logical_if_lock;

#ifdef CONFIG_PPA_QOS_8_VAP_SUPPORT
#define MAX_VAP_QOS_SUPPORT 8
#else
#define MAX_VAP_QOS_SUPPORT 4
#endif
#endif /*CONFIG_PPA_VAP_QOS_SUPPORT*/

extern QOS_QUEUE_LIST_ITEM *g_qos_queue;
extern PPA_LOCK g_qos_queue_lock;

#endif /*END OF CONFIG_PPA_QOS*/

extern QOS_QUEUE_LIST_ITEM *ppa_qos_queue_alloc_item(void);
extern void ppa_qos_queue_free_item(QOS_QUEUE_LIST_ITEM *obj);
extern void ppa_qos_queue_lock_list(void);
extern void ppa_qos_queue_unlock_list(void);
extern void __ppa_qos_queue_add_item(QOS_QUEUE_LIST_ITEM *obj);
extern void ppa_qos_queue_remove_item(int32_t q_num, PPA_IFNAME ifname[16], QOS_QUEUE_LIST_ITEM **pp_info);
extern void ppa_qos_queue_free_list(void);
extern int32_t ppa_qos_queue_lookup(int32_t q_num, PPA_IFNAME ifname[16], QOS_QUEUE_LIST_ITEM **pp_info);

#if defined(CONFIG_PPA_VAP_QOS_SUPPORT) && defined(CONFIG_SOC_GRX500)
extern QOS_LOGICAL_IF_LIST_ITEM *ppa_qos_logical_if_alloc_item(void);
extern void ppa_qos_logical_if_free_item(QOS_LOGICAL_IF_LIST_ITEM *obj);
extern void ppa_qos_logical_if_lock_list(void);
extern void ppa_qos_logical_if_unlock_list(void);
extern void __ppa_qos_logical_if_add_item(QOS_LOGICAL_IF_LIST_ITEM *obj);
extern void ppa_qos_logical_if_remove_item(PPA_IFNAME ifname[16], QOS_LOGICAL_IF_LIST_ITEM **pp_info);
extern void ppa_qos_logical_if_free_list(void);
extern int32_t ppa_qos_logical_if_lookup(int32_t subifId, PPA_IFNAME ifname[16], QOS_LOGICAL_IF_LIST_ITEM **pp_info);
extern int32_t ppa_set_logical_if_qos_classifier(PPA_CLASS_RULE *rule);
#endif // end of CONFIG_PPA_VAP_QOS_SUPPORT
#endif //end of __PPA_API_MFE_H__20100427_1703

